"use strict";

exports.__esModule = true;

var _lodash = require("lodash");

var _lodash2 = _interopRequireDefault(_lodash);

var _sqlTokenTypes = require("./sqlTokenTypes");

var _sqlTokenTypes2 = _interopRequireDefault(_sqlTokenTypes);

var _Indentation = require("./Indentation");

var _Indentation2 = _interopRequireDefault(_Indentation);

var _InlineBlock = require("./InlineBlock");

var _InlineBlock2 = _interopRequireDefault(_InlineBlock);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var SqlFormatter = function () {
    /**
     * @param {SqlTokenizer} tokenizer
     */
    function SqlFormatter(tokenizer) {
        _classCallCheck(this, SqlFormatter);

        this.tokenizer = tokenizer;
        this.indentation = new _Indentation2["default"]();
        this.inlineBlock = new _InlineBlock2["default"]();
        this.previousReservedWord = {};
    }

    /**
     * Format the whitespace in a SQL string to make it easier to read.
     *
     * @param {String} query The SQL query string
     * @return {String} formatted query
     */


    SqlFormatter.prototype.format = function format(query) {
        var tokens = this.tokenizer.tokenize(query);
        var formattedQuery = this.getFormattedQueryFromTokens(tokens);

        return formattedQuery.trim() + "\n";
    };

    SqlFormatter.prototype.getFormattedQueryFromTokens = function getFormattedQueryFromTokens(tokens) {
        var _this = this;

        var formattedQuery = "";

        tokens.forEach(function (token, index) {
            if (token.type === _sqlTokenTypes2["default"].WHITESPACE) {
                return;
            } else if (token.type === _sqlTokenTypes2["default"].LINE_COMMENT) {
                formattedQuery = _this.formatLineComment(token, formattedQuery);
            } else if (token.type === _sqlTokenTypes2["default"].BLOCK_COMMENT) {
                formattedQuery = _this.formatBlockComment(token, formattedQuery);
            } else if (token.type === _sqlTokenTypes2["default"].RESERVED_TOPLEVEL) {
                formattedQuery = _this.formatToplevelReservedWord(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _sqlTokenTypes2["default"].RESERVED_NEWLINE) {
                formattedQuery = _this.formatNewlineReservedWord(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _sqlTokenTypes2["default"].RESERVED) {
                formattedQuery = _this.formatWithSpaces(token, formattedQuery);
                _this.previousReservedWord = token;
            } else if (token.type === _sqlTokenTypes2["default"].OPEN_PAREN) {
                formattedQuery = _this.formatOpeningParentheses(tokens, index, formattedQuery);
            } else if (token.type === _sqlTokenTypes2["default"].CLOSE_PAREN) {
                formattedQuery = _this.formatClosingParentheses(token, formattedQuery);
            } else if (token.value === ",") {
                formattedQuery = _this.formatComma(token, formattedQuery);
            } else if (token.value === ":") {
                formattedQuery = _this.formatWithSpaceAfter(token, formattedQuery);
            } else if (token.value === "." || token.value === ";") {
                formattedQuery = _this.formatWithoutSpaces(token, formattedQuery);
            } else {
                formattedQuery = _this.formatWithSpaces(token, formattedQuery);
            }
        });
        return formattedQuery;
    };

    SqlFormatter.prototype.formatLineComment = function formatLineComment(token, query) {
        return this.addNewline(query + token.value);
    };

    SqlFormatter.prototype.formatBlockComment = function formatBlockComment(token, query) {
        return this.addNewline(this.addNewline(query) + this.indentComment(token.value));
    };

    SqlFormatter.prototype.indentComment = function indentComment(comment) {
        return comment.replace(/\n/g, "\n" + this.indentation.getIndent());
    };

    SqlFormatter.prototype.formatToplevelReservedWord = function formatToplevelReservedWord(token, query) {
        this.indentation.decreaseTopLevel();

        query = this.addNewline(query);

        this.indentation.increaseToplevel();

        query += this.equalizeWhitespace(token.value);
        return this.addNewline(query);
    };

    SqlFormatter.prototype.formatNewlineReservedWord = function formatNewlineReservedWord(token, query) {
        return this.addNewline(query) + this.equalizeWhitespace(token.value) + " ";
    };

    // Replace any sequence of whitespace characters with single space


    SqlFormatter.prototype.equalizeWhitespace = function equalizeWhitespace(string) {
        return string.replace(/\s+/g, " ");
    };

    // Opening parentheses increase the block indent level and start a new line


    SqlFormatter.prototype.formatOpeningParentheses = function formatOpeningParentheses(tokens, index, query) {
        // Take out the preceding space unless there was whitespace there in the original query
        var previousToken = tokens[index - 1];
        if (previousToken && previousToken.type !== _sqlTokenTypes2["default"].WHITESPACE) {
            query = _lodash2["default"].trimEnd(query);
        }
        query += tokens[index].value;

        this.inlineBlock.beginIfPossible(tokens, index);

        if (!this.inlineBlock.isActive()) {
            this.indentation.increaseBlockLevel();
            query = this.addNewline(query);
        }
        return query;
    };

    // Closing parentheses decrease the block indent level


    SqlFormatter.prototype.formatClosingParentheses = function formatClosingParentheses(token, query) {
        if (this.inlineBlock.isActive()) {
            this.inlineBlock.end();
            return this.formatWithSpaceAfter(token, query);
        } else {
            this.indentation.decreaseBlockLevel();
            return this.formatWithSpaces(token, this.addNewline(query));
        }
    };

    // Commas start a new line (unless within inline parentheses or SQL "LIMIT" clause)


    SqlFormatter.prototype.formatComma = function formatComma(token, query) {
        query = _lodash2["default"].trimEnd(query) + token.value + " ";

        if (this.inlineBlock.isActive()) {
            return query;
        } else if (/^LIMIT$/i.test(this.previousReservedWord.value)) {
            return query;
        } else {
            return this.addNewline(query);
        }
    };

    SqlFormatter.prototype.formatWithSpaceAfter = function formatWithSpaceAfter(token, query) {
        return _lodash2["default"].trimEnd(query) + token.value + " ";
    };

    SqlFormatter.prototype.formatWithoutSpaces = function formatWithoutSpaces(token, query) {
        return _lodash2["default"].trimEnd(query) + token.value;
    };

    SqlFormatter.prototype.formatWithSpaces = function formatWithSpaces(token, query) {
        return query + token.value + " ";
    };

    SqlFormatter.prototype.addNewline = function addNewline(query) {
        return _lodash2["default"].trimEnd(query) + "\n" + this.indentation.getIndent();
    };

    return SqlFormatter;
}();

exports["default"] = SqlFormatter;
module.exports = exports["default"];