"use strict";

exports.__esModule = true;

var _lodash = require("lodash");

var _lodash2 = _interopRequireDefault(_lodash);

var _sqlTokenTypes = require("./sqlTokenTypes");

var _sqlTokenTypes2 = _interopRequireDefault(_sqlTokenTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var SqlTokenizer = function () {
    /**
     * @param {Object} cfg
     *  @param {String[]} cfg.reservedWords Reserved words in SQL
     *  @param {String[]} cfg.reservedToplevelWords Words that are set to new line and on first indent level
     *  @param {String[]} cfg.reservedNewlineWords Words that are set to newline
     *  @param {String[]} cfg.stringTypes String types to enable: "", '', ``, []
     *  @param {String[]} cfg.openParens Opening parentheses to enable, like (, [
     *  @param {String[]} cfg.closeParens Closing parentheses to enable, like ), ]
     *  @param {String[]} cfg.variableTypes Prefixes for variables, like @ and :
     */
    function SqlTokenizer(cfg) {
        _classCallCheck(this, SqlTokenizer);

        this.WORD_REGEX = /^(\w+)/;
        this.WHITESPACE_REGEX = /^(\s+)/;
        this.LINE_COMMENT_REGEX = /^((?:#|--).*?(?:\n|$))/;
        this.BLOCK_COMMENT_REGEX = /^(\/\*[^]*?(?:\*\/|$))/;
        this.NUMBER_REGEX = /^((-\s*)?[0-9]+(\.[0-9]+)?|0x[0-9a-fA-F]+|0b[01]+)\b/;
        this.OPERATOR_REGEX = /^(!=|<>|==|<=|>=|!<|!>|\|\||.)/;

        this.RESERVED_TOPLEVEL_REGEX = this.createReservedWordRegex(cfg.reservedToplevelWords);
        this.RESERVED_NEWLINE_REGEX = this.createReservedWordRegex(cfg.reservedNewlineWords);
        this.RESERVED_PLAIN_REGEX = this.createReservedWordRegex(cfg.reservedWords);

        this.STRING_REGEX = this.createStringRegex(cfg.stringTypes);

        this.OPEN_PAREN_REGEX = this.createParenRegex(cfg.openParens);
        this.CLOSE_PAREN_REGEX = this.createParenRegex(cfg.closeParens);

        this.PLAIN_VARIABLE_REGEX = this.createVariableRegex(cfg.variableTypes, "[a-zA-Z0-9._$]+");
        this.QUOTED_VARIABLE_REGEX = this.createVariableRegex(cfg.variableTypes, this.createStringPattern(cfg.stringTypes));
    }

    SqlTokenizer.prototype.createReservedWordRegex = function createReservedWordRegex(reservedWords) {
        var reservedWordsPattern = reservedWords.join("|").replace(/ /g, "\\s+");
        return new RegExp("^(" + reservedWordsPattern + ")\\b", "i");
    };

    SqlTokenizer.prototype.createStringRegex = function createStringRegex(stringTypes) {
        return new RegExp("^(" + this.createStringPattern(stringTypes) + ")");
    };

    // This enables the following string patterns:
    // 1. backtick quoted string using `` to escape
    // 2. square bracket quoted string (SQL Server) using ]] to escape
    // 3. double quoted string using "" or \" to escape
    // 4. single quoted string using '' or \' to escape


    SqlTokenizer.prototype.createStringPattern = function createStringPattern(stringTypes) {
        var patterns = {
            "\"\"": "((\"[^\"\\\\]*(?:\\\\.[^\"\\\\]*)*(\"|$))+)",
            "''": "(('[^'\\\\]*(?:\\\\.[^'\\\\]*)*('|$))+)",
            "``": "((`[^`]*($|`))+)",
            "[]": "((\\[[^\\]]*($|\\]))(\\][^\\]]*($|\\]))*)"
        };

        return stringTypes.map(function (t) {
            return patterns[t];
        }).join("|");
    };

    SqlTokenizer.prototype.createParenRegex = function createParenRegex(parens) {
        return new RegExp("^(" + parens.map(function (p) {
            return _lodash2["default"].escapeRegExp(p);
        }).join("|") + ")");
    };

    SqlTokenizer.prototype.createVariableRegex = function createVariableRegex(variableTypes, pattern) {
        if (variableTypes.length === 0) {
            return false;
        }
        return new RegExp("^((?:" + variableTypes.map(_lodash2["default"].escapeRegExp).join("|") + ")(?:" + pattern + "))\\S");
    };

    /**
     * Takes a SQL string and breaks it into tokens.
     * Each token is an object with type and value.
     *
     * @param {String} input The SQL string
     * @return {Object[]} tokens An array of tokens.
     *  @return {String} token.type
     *  @return {String} token.value
     */


    SqlTokenizer.prototype.tokenize = function tokenize(input) {
        var tokens = [];
        var token = void 0;

        // Keep processing the string until it is empty
        while (input.length) {
            // Get the next token and the token type
            token = this.getNextToken(input, token);
            // Advance the string
            input = input.substring(token.value.length);

            tokens.push(token);
        }
        return tokens;
    };

    SqlTokenizer.prototype.getNextToken = function getNextToken(input, previousToken) {
        return this.getWhitespaceToken(input) || this.getCommentToken(input) || this.getStringToken(input) || this.getOpenParenToken(input) || this.getCloseParenToken(input) || this.getVariableToken(input) || this.getNumberToken(input) || this.getReservedWordToken(input, previousToken) || this.getWordToken(input) || this.getOperatorToken(input);
    };

    SqlTokenizer.prototype.getWhitespaceToken = function getWhitespaceToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].WHITESPACE,
            regex: this.WHITESPACE_REGEX
        });
    };

    SqlTokenizer.prototype.getCommentToken = function getCommentToken(input) {
        return this.getLineCommentToken(input) || this.getBlockCommentToken(input);
    };

    SqlTokenizer.prototype.getLineCommentToken = function getLineCommentToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].LINE_COMMENT,
            regex: this.LINE_COMMENT_REGEX
        });
    };

    SqlTokenizer.prototype.getBlockCommentToken = function getBlockCommentToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].BLOCK_COMMENT,
            regex: this.BLOCK_COMMENT_REGEX
        });
    };

    SqlTokenizer.prototype.getStringToken = function getStringToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].STRING,
            regex: this.STRING_REGEX
        });
    };

    SqlTokenizer.prototype.getOpenParenToken = function getOpenParenToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].OPEN_PAREN,
            regex: this.OPEN_PAREN_REGEX
        });
    };

    SqlTokenizer.prototype.getCloseParenToken = function getCloseParenToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].CLOSE_PAREN,
            regex: this.CLOSE_PAREN_REGEX
        });
    };

    SqlTokenizer.prototype.getVariableToken = function getVariableToken(input) {
        return this.getPlainVariableToken(input) || this.getQuotedVariableToken(input);
    };

    SqlTokenizer.prototype.getPlainVariableToken = function getPlainVariableToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].VARIABLE,
            regex: this.PLAIN_VARIABLE_REGEX
        });
    };

    SqlTokenizer.prototype.getQuotedVariableToken = function getQuotedVariableToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].VARIABLE,
            regex: this.QUOTED_VARIABLE_REGEX
        });
    };

    // Decimal, binary, or hex numbers


    SqlTokenizer.prototype.getNumberToken = function getNumberToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].NUMBER,
            regex: this.NUMBER_REGEX
        });
    };

    // Punctuation and symbols


    SqlTokenizer.prototype.getOperatorToken = function getOperatorToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].OPERATOR,
            regex: this.OPERATOR_REGEX
        });
    };

    SqlTokenizer.prototype.getReservedWordToken = function getReservedWordToken(input, previousToken) {
        // A reserved word cannot be preceded by a "."
        // this makes it so in "mytable.from", "from" is not considered a reserved word
        if (previousToken && previousToken.value && previousToken.value === ".") {
            return;
        }
        return this.getToplevelReservedToken(input) || this.getNewlineReservedToken(input) || this.getPlainReservedToken(input);
    };

    SqlTokenizer.prototype.getToplevelReservedToken = function getToplevelReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].RESERVED_TOPLEVEL,
            regex: this.RESERVED_TOPLEVEL_REGEX
        });
    };

    SqlTokenizer.prototype.getNewlineReservedToken = function getNewlineReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].RESERVED_NEWLINE,
            regex: this.RESERVED_NEWLINE_REGEX
        });
    };

    SqlTokenizer.prototype.getPlainReservedToken = function getPlainReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].RESERVED,
            regex: this.RESERVED_PLAIN_REGEX
        });
    };

    SqlTokenizer.prototype.getWordToken = function getWordToken(input) {
        return this.getTokenOnFirstMatch({
            input: input,
            type: _sqlTokenTypes2["default"].WORD,
            regex: this.WORD_REGEX
        });
    };

    SqlTokenizer.prototype.getTokenOnFirstMatch = function getTokenOnFirstMatch(_ref) {
        var input = _ref.input;
        var type = _ref.type;
        var regex = _ref.regex;

        var matches = input.match(regex);

        if (matches) {
            return { type: type, value: matches[1] };
        }
    };

    return SqlTokenizer;
}();

exports["default"] = SqlTokenizer;
module.exports = exports["default"];